package CUPS::Backend;

use strict;
use warnings;
use File::Basename;
use POSIX qw( :fcntl_h );
use Fcntl qw( :DEFAULT );
use Symbol;
use URI;

our $VERSION = '0.02';

######################################################
sub new
{
    my( $package, $args ) = @_;
    my $self = bless {}, $package;
    $self->init( $args );
    return $self;
}

######################################################
sub init
{
    my( $self, $args ) = @_;
    $self->{name} = $args->{name} || ref $self;
    $self->{type} = $args->{type} || 'direct';
    $self->{backend} = $args->{backend};
    unless( $self->{backend} ) {
        $self->{backend} = basename $0;
    }
}


######################################################
sub _msg
{
    my( $self, $prefix, $msg ) = @_;
    $prefix .= ": ";
    $prefix .= "Job $self->{jobid} " if $self->{jobid};
    $msg .= "\n" unless $msg =~ /\n$/;
    $msg =~ s/^/$prefix/gm;
    return $msg;
}

sub error
{
    my( $self, @msg ) = @_;
    print STDERR $self->_msg( ERROR => join '', @msg );
}

sub info
{
    my( $self, @msg ) = @_;
    print STDERR $self->_msg( INFO => join '', @msg );
}

sub page
{
    my( $self, $page, $copies ) = @_;
    print STDERR "PAGE: $page $copies\n";
}

sub output
{
    my( $self, @msg ) = @_;
    print join '', @msg;
}
        
######################################################
sub information
{
    my( $self ) = @_;
    $self->output( qq($self->{type} $self->{backend} "Unknown" "$self->{name}"\n) );
}

######################################################
sub cmdline
{
    my( $self, $argv ) = @_;
    $argv ||= \@ARGV;
    if( 0 == @$argv ) {
        $self->information;
        exit 0;
    }

    if( 1 == @$argv and $argv->[0] eq '--desc' ) {
        $self->output( qq(CUPS backend logs to a YAML file\n) );
        exit 0;
    }

    unless( 5==@$argv or 6==@$argv ) {
        die "Usage: $0 job-id user title copies options [file]\n";
    }

    @{ $self }{ qw( jobid user title copies options_str input ) } = @$argv;

    if( 5==@$argv ) {
        $self->{copies} = 1;
        $self->{input} = '-';
    }
    else {
        $self->{copies} = 0+$argv->[3];
    }
    $self->parse_options;
    return 1;
}

######################################################
sub parse_uri
{
    my( $self, $uri ) = @_;
    $uri ||= $ENV{DEVICE_URI};
    return unless $uri;

    $self->{uri} = URI->new( $uri );
    die "INFO: Unable to parse DEVICE_URI: $uri\n" unless $self->{uri};
    if( $self->{uri}->scheme ne $self->{backend} ) {
        $self->info( "Scheme (", $self->{uri}->scheme, ") differs from backend name ($self->{backend})" );
    }
    return $self->{uri};
}

######################################################
sub parse_env
{
    my( $self, $env ) = @_;
    $env ||= \%ENV;
    foreach my $key ( qw( CHARSET CONTENT_TYPE DEVICE_URI IPP_PORT
                            FINAL_CONTENT_TYPE PPD PRINTER RIP_MAX_CACHE
                            SERVER_ADMIN SOFTWARE TMPDIR USER
                    ) ) {
        next unless defined $env->{$key};
        $self->{ENV}{$key} = $env->{$key};
    }

    foreach my $key ( keys %$env ) {
        my $nkey = $key;
        next unless $nkey =~ s/CUPS_//;
        $self->{ENV}{$key} = $env->{$key};
        $self->{ENV}{$nkey} = $env->{$key};
    }
    if( $env->{DEVICE_URI} ) {
        $self->parse_uri( $env->{DEVICE_URI} );
    }
}

sub env
{
    my( $self, $opt, $value ) = @_;
    my $ret = $self->{ENV}{$opt};
    if( 3==@_ ) {
        $self->{ENV}{$opt} = $value;
        if( $opt =~ s/CUPS_// ) {
            $self->{ENV}{$opt} = $value;
        }
        elsif( exists $self->{ENV}{ "CUPS_$opt" } ) {
            $self->{ENV}{ "CUPS_$opt" } = $value;
        }
    }
    return $ret;
}

######################################################
sub parse_options
{
    my( $self, $options ) = @_;
    $options ||= $self->{options_str};

    my $opts = $self->{options} = {};
    foreach my $opt ( split ' ', $options ) {
        if( $opt =~ /^(.+?)=(.+)$/ ) {
            $opts->{$1} = $2;
        }
        elsif( $opt =~ /^no(.+)$/ ) {
            $opts->{$1} = 0;
        }
        elsif( $opt =~ /^(.+)$/ ) {
            $opts->{$opt} = 1;
        }
    }
}

######################################################
sub option
{
    my( $self, $opt, $value ) = @_;
    my $ret = $self->{options}{$opt};
    $self->{options}{$opt} = $value if 3==@_;
    return $ret;
}

######################################################
sub lp_options
{
    my( $self ) = @_;
    my @opt;
    foreach my $key ( sort keys %{$self->{options}} ) {
        if( '0' eq $self->{options}{$key} ) {
            push @opt, "no$key";
        }
        elsif ( '1' eq $self->{options}{$key} ) {
            push @opt, $key;
        }
        else {
            push @opt, join '=', $key, quotemeta $self->{options}{$key};
        }
    }
    return join ' ', map { "-o $_" } @opt;
}

######################################################
# Accessors
sub jobid { $_[0]->{jobid} }
sub user { $_[0]->{user} }
sub title { $_[0]->{title} }
sub copies { $_[0]->{copies} }
sub input { $_[0]->{input} }
sub options { $_[0]->{options_str} }

sub uri { $_[0]->{uri} }
sub scheme { $_[0]->uri->scheme }
sub opaque { $_[0]->uri->opaque }
sub path { $_[0]->uri->path }

######################################################
sub slurp
{
    my( $self, $fh ) = @_;

    my( $ret, $buffer );
    $buffer = '';
    my $in;
    if( $self->{input} eq '-' ) {
        $self->info( "Reading STDIN" );
        $in = *STDIN;
    }
    else {
        $in = gensym;
        $self->info( "Reading $self->{input}" );
        sysopen( $in, $self->{input}, O_RDONLY )
                 or die "Unable to read $self->{input}: $!";
    }
    binmode( $in );
    binmode( $fh ) if $fh;
    my( $n, $size, $bsize, $ok ) = ( 0, 0, 1024*1024, 1 );
    while( $ok and $n = sysread( $in, $buffer, $bsize ) ) {
        $self->info( "Read $n bytes" );
        if( $fh ) {
            my $n2 = syswrite( $fh, $buffer, $n );
            $self->info( "Wrote $n2 bytes" );
            $size += $n2;
        }
        else {
            $size += $n;
            $ret .= $buffer;
        }
        if( $n < $bsize ) {
            # $self->info( "Short read" );
            if( $buffer =~ s/\x1A\x00+\x0D$// ) {
                # $ok = 0;
                $self->info( "Contains ^Z" );
            }
        }
    }
    return $size if $fh;
    return( $ret, $size ) if wantarray;
    return $ret;
}

1;
__END__

=head1 NAME

CUPS::Backend - Helper routines for writing CUPS backends

=head1 SYNOPSIS

  use CUPS::Backend;

=head1 DESCRIPTION

=head2 EXPORT


=head1 SEE ALSO

L<Net::CUPS>

=head1 AUTHOR

Philip Gwyn, E<lt>gwyn-at-cpan.orgE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2008 by Philip Gwyn

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.8 or,
at your option, any later version of Perl 5 you may have available.


=cut
