package CUPS::Spool;

use strict;
use warnings;
use File::Basename;
use File::Spec;
use Symbol;
use URI;
use Errno qw( EBADF );

use enum qw(
        :IPP_TAG_=0 
            ZERO OPERATION JOB END PRINTER
            UNSUPPORTED_GROUP SUBSCRIPTION EVENT_NOTIFICATION
            UNSUPPORTED_VALUE=0x10 
            DEFAULT UNKNOWN NOVALUE
            NOTSETTABLE=0x15 
            DELETEATTR ADMINDEFINE
            INTEGER=0x21
            BOOLEAN ENUM 
            STRING=0x30 DATE RESOLUTION RANGE BEGIN_COLLECTION
            TEXTLANG NAMELANG END_COLLECTION
            TEXT=0x41 NAME 
            KEYWORD=0x44 URI
            URISCHEME CHARSET LANGUAGE MIMETYPE MEMBERNAME
            MASK=0x7fffffff COPY=0x80000000
        :IPP_JOB_=3 
            PENDING HELD 
            PROCESSING STOPPED CANCELLED ABORTED COMPLETED
    );

our $VERSION='0.01';

######################################################
sub new
{
    my( $package, $args )=@_;
    my $self = bless {}, $package;
    $self->init( $args );
    return $self;
}

######################################################
sub init
{
    my( $self, $args )=@_;
    $self->{dir} = $args->{dir} || '/var/spool/cups';
}

######################################################
sub list
{
    my( $self ) = @_;
    my @ret;
    opendir DIR, $self->{dir} or die "Unable to read $self->{dir}: $!\n";
    while( my $file = readdir DIR ) {
        next unless $file =~ /^c(\d+)$/;
        push @ret, 0+$1;
    }
    closedir DIR;
    return sort { $a <=> $b } @ret;
}

######################################################
sub control_file
{
    my( $self, $num ) = @_;
    return File::Spec->catfile( $self->{dir}, sprintf "c%05i", 0+$num );
}

sub data_file
{
    my( $self, $num, $n ) = @_;
    return unless 0+$n;
    return File::Spec->catfile( $self->{dir}, 
                                sprintf "d%05i-%03i", 0+$num, 0+$n
                              );
}

######################################################
sub load
{
    my( $self, $num ) = @_;

    my( $control, $load_file );
    if( -f $num ) {
        $control = $num;
        die "No number in $control" unless $num =~ /(\d+)$/;
        $num = 0+$1;
        $load_file = 1;
    }
    else {
        $control = $self->control_file( $num );
    }
    return unless -f $control and -r _;
    open FILE, $control;
    binmode( FILE );
    my $buffer = eval { local $/; <FILE> };
    close FILE;

    my %data = ( control => $control, number => sprintf "%05i", $num );

    @data{ qw( version_major version_minor op_status request_id ) }
                = unpack( "CCnN", $buffer );

    $data{version} = "$data{version_major}.$data{version_minor}";
    $data{version}+=0;

    unless( $data{version_major} == 1 ) {
        $! = EBADF;
        return;
    }

    my $len = length $buffer;
    my $offset = 8;
    my( $group, @attr );
    
    TAG:
    while( $offset < $len ) {
        # [tag]
        my $tag = unpack "C", substr( $buffer, $offset++ );
        if( $tag < IPP_TAG_UNSUPPORTED_VALUE ) {
            $group = $tag;
            next TAG;
        }
        # LL [name]
        my $len = unpack "n", substr( $buffer, $offset ); $offset+=2;
        my $name = substr( $buffer, $offset, $len ); $offset += $len;

        # LL [value]
        $len = unpack "n", substr( $buffer, $offset ); $offset+=2;
        my $data = substr( $buffer, $offset, $len ); $offset+=$len;
        my $value;
        if( $tag == IPP_TAG_INTEGER or $tag == IPP_TAG_ENUM ) {
            $value = unpack "N", $data;
        }
        elsif( $tag == IPP_TAG_BOOLEAN ) {
            $value = unpack "C", $data;
        }
        elsif( $tag == IPP_TAG_TEXT or $tag == IPP_TAG_NAME or
               $tag == IPP_TAG_KEYWORD or $tag == IPP_TAG_STRING
               or $tag == IPP_TAG_URI  or $tag == IPP_TAG_URISCHEME
               or $tag == IPP_TAG_CHARSET or $tag == IPP_TAG_LANGUAGE
               or $tag == IPP_TAG_MIMETYPE ) {
            $value = $data;
        }
        elsif( $tag == IPP_TAG_DATE ) {
            $value = $data;
        }
        elsif( $tag == IPP_TAG_RESOLUTION ) {
            $value = {  xres => unpack( "N", $data ),
                        yres => unpack( "N", substr( $data, 4 ) ),
                        units => unpack( "C", substr( $data, 8 ) )
                     };
        }
        elsif( $tag == IPP_TAG_RANGE ) {
            $value = {  lower => unpack( "N", $data ),
                        upper => unpack( "N", substr( $data, 4 ) )
                     };
        }
        elsif( $tag == IPP_TAG_TEXTLANG or $tag == IPP_TAG_NAMELANG ) {
            my $len1 = unpack "n", $data;
            my $len2 = unpack "n", substr( $data, 2+$len1 );
            $value = {  charset => substr( $data, 2, $len1 ),
                        text    => substr( $data, 2+$len1, $len2 )
                     };
        }
        elsif( $tag == IPP_TAG_BEGIN_COLLECTION or 
               $tag == IPP_TAG_MEMBERNAME ) {
            die "IPP collections not supported";
        }
        elsif( $tag == IPP_TAG_END ) {
            last TAG;
        }
        else {
            $value = $data;
        }

        if( $name eq '' ) {
            
            my $A = $attr[-1];
            $A->{value} = [ $A->{value} ] unless 'ARRAY' eq ref $A->{value};
            push @{ $A->{value} }, $value;
        }
        else {
            push @attr, { name => $name, 
                          tag => $tag, 
                          value => $value, 
                          group => $group 
                        };
        }
    }

    $data{attr} = \@attr;

    my @data;
    {
        local $!;
        local $self->{dir} = dirname $control;
        my $N = 1;
        while ( 1 ) {
            my $data = $self->data_file( $num, $N );
            last unless -f $data;
            push @data, $data;
            $N++;
        }
    }
    
    $data{data} = \@data;
    return $self->build_control( \%data );

}

######################################################
sub build_control
{
    my( $self, $parms ) = @_;
    return CUPS::Spool::Control->new( $parms );
}

#############################################################################
package CUPS::Spool::Control;

use strict;
use warnings;

use Errno qw( ENOENT );

######################################################
sub new
{
    my( $package, $args ) = @_;
    my $self = bless {}, $package;
    $self->init( $args );
    return $self;
}

######################################################
sub init
{
    my( $self, $args ) = @_;
    @{ $self }{ keys %$args } = values %$args;
    $self->{_attrs} = [];
    foreach my $A ( @{ $self->{attr}||[] } ) {
        next if exists $self->{_attr}{ $A->{name} };
        $self->{_attr}{ $A->{name} } = $A;
    }
}

######################################################
sub number          { $_[0]->{number} }
sub version         { $_[0]->{version} }
sub version_major   { $_[0]->{version_major} }
sub version_minor   { $_[0]->{version_minor} }
sub op_status       { $_[0]->{op_status} }
sub request_id      { $_[0]->{request_id} }
sub control_file    { $_[0]->{control} }
sub data_files
{ 
    my( $self ) = @_;
    return @{ $self->{data} } if wantarray;
    return unless @{ $self->{data} };
    return $self->{data};
}

######################################################
sub PENDING { $_[0]->{_attr}{'job-state'}{value} == CUPS::Spool::IPP_JOB_PENDING() }
sub HELD { $_[0]->{_attr}{'job-state'}{value} == CUPS::Spool::IPP_JOB_HELD() }
sub PROCESSING { $_[0]->{_attr}{'job-state'}{value} == CUPS::Spool::IPP_JOB_PROCESSING() }
sub STOPPED { $_[0]->{_attr}{'job-state'}{value} == CUPS::Spool::IPP_JOB_STOPPED() }
sub CANCELLED { $_[0]->{_attr}{'job-state'}{value} == CUPS::Spool::IPP_JOB_CANCELLED() }
sub ABORTED { $_[0]->{_attr}{'job-state'}{value} == CUPS::Spool::IPP_JOB_ABORTED() }
sub COMPLETED { $_[0]->{_attr}{'job-state'}{value} == CUPS::Spool::IPP_JOB_COMPLETED() }


######################################################
sub attr
{
    my( $self, $name ) = @_;
    if( 1 == @_ ) {
        return map { $self->build_attr( $_ ) } @{ $self->{attr} };
    }

    unless( $self->{_attr}{$name} ) {
        $! = ENOENT;
        return;
    }
    return $self->build_attr( $self->{_attr}{$name} );
}

sub build_attr
{
    my( $self, $A ) = @_;
    return CUPS::Spool::Control::Attr->new( $A );
}

#############################################################################
package CUPS::Spool::Control::Attr;

use strict;
use warnings;

######################################################
sub new
{
    my( $package, $args ) = @_;
    my $self = bless {}, $package;
    $self->init( $args );
    return $self;
}

######################################################
sub init
{
    my( $self, $args ) = @_;
    @{ $self }{ keys %$args } = values %$args;
}

######################################################
sub group   { $_[0]->{group} }
sub tag     { $_[0]->{tag} }
sub name    { $_[0]->{name} }
sub value   { $_[0]->{value} }


1;
__END__

=head1 NAME

CUPS::Spool - Read the CUPS spool files

=head1 SYNOPSIS

  use CUPS::Spool;

=head1 DESCRIPTION

=head2 EXPORT


=head1 SEE ALSO

L<Net::CUPS>

=head1 AUTHOR

Philip Gwyn E<lt>gwyn-at-cpan.orgE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2008-2009 by Philip Gwyn

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself either Perl version 5.8.8 or
at your option,  any later version of Perl 5 you may have available.


=cut
